#ifndef _TRANSLATOR_
#define _TRANSLATOR_

#include <iostream>
#include <deque>
#include <set>
#include <vector>
#include <algorithm>
#include <queue>
#include <list>

#include "./CoverageHypothesis.h"
#include "./LexicalHypothesis.h"

#include "../Model/LatticeManager.h"
#include "../Model/ReorderingModel.h"

#include "../DataStructure/ObjectPool.h"

using namespace std;

typedef set<CoverageHypothesis>           CoverageHypothesisMap;
typedef vector<const CoverageHypothesis*> CoverageHypothesisStack;
typedef vector<CoverageHypothesisStack>   CoverageHypothesisStackArray;

class LexicalHypothesis;
// FixedObjectPool<LexicalHypothesis> lexHypPool;

class CovHypCmpByScoreSoFar{
 public:
  inline bool operator()(const CoverageHypothesis* a, const CoverageHypothesis* b) const {
    return a->getMaxScore() > b->getMaxScore();
    // return a->getMaxScoreSoFar() > b->getMaxScoreSoFar();
  }
};

class Translator{
private:
  const Parameter&      param;
  LanguageModel&        lm;

  float maxScore;
  float minScore;
  float beamThreshold;
  unsigned beamWidth;

  const unsigned                 sentenceLength;
  const LatticeManager&          ltcManager;
  CoverageHypothesisStackArray   stackArray;
  CoverageHypothesisMap          covHypoMap[maxSentenceLength+1];

  unsigned lexHypNum;

  void initialize(){
    lexHypPool.clear();

    Coverage emptyCoverage;
    CoverageHypothesis& covHypo = const_cast<CoverageHypothesis&>(
        *(covHypoMap[0].insert(CoverageHypothesis(emptyCoverage)).first)
    );
    stackArray[0].push_back( &covHypo );
    LexicalHypothesis emptyHypo(&covHypo.getCoverage(), 0.0, 0.0);
    LexicalHypothesis* ptrEmptyHyp = lexHypPool.pop();
    *ptrEmptyHyp = emptyHypo;
    covHypo.push( ptrEmptyHyp );

    for(unsigned cardinality=1; cardinality<=sentenceLength; cardinality++)
      stackArray[cardinality].reserve(beamWidth);
  }

  const string& convertHypoArrayToString(const deque<const LexicalHypothesis*>& lexHypoArray) const{
    static string str = "";
    str.clear();

    for(unsigned i=0; i<lexHypoArray.size(); i++)
      str += lexHypoArray[i]->getLastPhrasePair()->getTgtPhrase().toString();

    return str;
  }

  const float computeReorderingScore(const PhrasePair* const predPP, const PhrasePair* const succPP, const ReorderingLabel reorderingLabel) const {
    float reorderingScore = 0.0F;

    float predecessorScore = 0.0F;
    switch(reorderingLabel){
      case MONO_LABEL:        // monotone
        predecessorScore = (predPP != NULL) ? predPP->getWeightedRMScore()[S_MONO] : 0.0F;
        reorderingScore = predecessorScore + succPP->getWeightedRMScore()[P_MONO];
        break;
      case SWAP_LABEL:       // swap
        reorderingScore = predPP->getWeightedRMScore()[S_SWAP] + succPP->getWeightedRMScore()[P_SWAP];
        break;
      default: // discont.
        predecessorScore = (predPP != NULL) ? predPP->getWeightedRMScore()[S_DISCONT] : 0.0F;
        reorderingScore = predecessorScore + succPP->getWeightedRMScore()[P_DISCONT];
    }

    return reorderingScore;
  }

  const float computeSentenceEndScore(const PhrasePair& lastPP, const unsigned end) const {
    const float lmScore         = param.getWeightL() * lm.getLMScoreForSentenceEnd();
    const float distPenalty     = param.getWeightD()[DIST_PENALTY] * (-ReorderingModel::calcDistortionPenalty(end, sentenceLength));
    const bool  monotonicity    = (end == sentenceLength-1);
    const float reorderingScore = monotonicity ? lastPP.getWeightedRMScore()[S_MONO] : lastPP.getWeightedRMScore()[S_DISCONT];

    const float sentenceEndScore = lmScore + distPenalty + reorderingScore;
    return sentenceEndScore;
  }

  void updateMaxScore(const float score){
    maxScore = score;
    minScore = std::max(minScore, maxScore + beamThreshold);
  }

  bool isTooBad(const float& score, const CoverageHypothesis& covHypo) const {
    return score < minScore;
    // return (score < minScore || covHypo.isTooBad(score));
  }

  bool isTooBad(const float& score, const CoverageHypothesis* dummy) const {
    return score < minScore;
  }

  void pruneCardinality(LexicalHypothesis** lexHypStack, unsigned& lexHypNumOfCardinality, CoverageHypothesisMap& coverageHypoMap){
    // calc threshold score by stack size
    priority_queue<float> pq;
    for(LexicalHypothesis** iter = lexHypStack; iter != &lexHypStack[lexHypNumOfCardinality]; iter++)
      pq.push( (*iter)->getEstimatedScore() );
    for(unsigned i=0; i<beamWidth; i++)
      pq.pop();
    minScore = pq.top();

    // pruning for cardinality stack
    LexicalHypothesis** end = lexHypStack + lexHypNumOfCardinality;
    lexHypNumOfCardinality = 0;
    for(LexicalHypothesis** iter = lexHypStack; iter < end; iter++)
      if( (*iter)->getEstimatedScore() >= minScore )
	lexHypStack[lexHypNumOfCardinality++] = *iter;

    // pruning for coverage stack
    CoverageHypothesisMap::iterator iter = coverageHypoMap.begin();
    while( iter != coverageHypoMap.end() ){
      if(iter->getStackSize())
	(const_cast<CoverageHypothesis&>(*iter)).prune(minScore);
      iter++;
    }

  }

  void pruneCardinalityLazy(const unsigned cardinality, const LexicalHypothesis** lexHypStack, unsigned& lexHypNumOfCardinality, CoverageHypothesisMap& coverageHypoMap){
    CoverageHypothesisStack& currentStack = stackArray[cardinality];

    if(lexHypNumOfCardinality > beamWidth){
      priority_queue<float> pq;
      const LexicalHypothesis** const end = lexHypStack + lexHypNumOfCardinality;
      for(const LexicalHypothesis** iter = lexHypStack; iter != end; iter++)
	pq.push((*iter)->getEstimatedScore());
      for(unsigned i=0; i<beamWidth; i++) pq.pop();
      minScore = pq.top();
    }

    CoverageHypothesisMap::const_iterator iter = coverageHypoMap.begin();
    while( iter != coverageHypoMap.end() ){
      if(iter->getMaxScore() < minScore)
	coverageHypoMap.erase(iter++);
      else {
        const_cast<CoverageHypothesis&>(*iter).prune(minScore);
        currentStack.push_back(&(*iter));
	iter++;
      }
    }

    sort(currentStack.begin(), currentStack.end(), CovHypCmpByScoreSoFar());
  }

  /*
  void pruneCardinality(const unsigned cardinality, const CoverageHypothesisMap& coverageHypoMap){
    CoverageHypothesisStack& currentStack = stackArray[cardinality];

    float threshold = minScore;

    if(coverageHypoMap.size() > stackSize){
      const float thresholdOfStackSize = getThresholdOfStackSize(coverageHypoMap.begin(), coverageHypoMap.end(), ::stackSize);
      if(threshold < thresholdOfStackSize) threshold = thresholdOfStackSize;
    }

    for(CoverageHypothesisMap::const_iterator iter = coverageHypoMap.begin();
      iter != coverageHypoMap.end();
      iter++)
      if(iter->getMaxScore() > threshold){
        const_cast<CoverageHypothesis&>(*iter).prune();
        currentStack.push_back(&(*iter));
      }
    sort(currentStack.begin(), currentStack.end(), hypoPtrScoreGreater<CoverageHypothesis>);
    while(currentStack.size() > stackSize) currentStack.pop_back();
  }
  */

  void expandCardinality(const unsigned cardinality){
    maxScore = -10000.0F;
    minScore = -20000.0F;
    unsigned lexHypNumOfCardinality = 0;
    LexicalHypothesis* lexHypStack[10000] = { NULL };

    CoverageHypothesisMap& coverageHypoMap = covHypoMap[cardinality];

    const unsigned curMaxPhraseLength = (maxPhraseLength > cardinality) ? cardinality : maxPhraseLength;
    // t[Y
    for(unsigned srcPhraseLength = 1; srcPhraseLength <= curMaxPhraseLength; srcPhraseLength++){
      // WJX^bNCoverageHypothesis
      unsigned prevCardinality = cardinality - srcPhraseLength;
      for(CoverageHypothesisStack::iterator covHypoIndex = stackArray[prevCardinality].begin();
          covHypoIndex != stackArray[prevCardinality].end();
          covHypoIndex++){
        const CoverageHypothesis& covHypo  = **covHypoIndex;
        const Coverage&           coverage = covHypo.getCoverage();

        // V|󂷂t[Y̊Jnʒu
        const unsigned startPosMax = sentenceLength - srcPhraseLength;
        for(unsigned start = 0; start <= startPosMax; start++){
          // V|󂷂t[Y̏I[ʒu
          unsigned end = start+srcPhraseLength-1;

          const Coverage& newCoveredCoverage = ltcManager.getCoverageLattice().get(end, start);
          // V|󂷂t[Yɖ|ς݂̒P܂ޏꍇ͎̌
          if( coverage.isOverlap(newCoveredCoverage) ) continue;

          const Coverage& newHypoCoverage = coverage | newCoveredCoverage;
	  CoverageHypothesis* newCovHypo = NULL;

	  // LexicalHypothesisWJ
         const unsigned size = covHypo.getStackSize();
          for(unsigned lexHypoIndex = 0; lexHypoIndex < size; lexHypoIndex++){
            const LexicalHypothesis& lexHypo = covHypo[lexHypoIndex]; // WJ

            // Õt[Y̊JnEI[ʒu
            const PhrasePair* const predPP = lexHypo.getLastPhrasePair();
            const int predEnd = lexHypo.getLastCoveredSrcPos();
            const int predBeg = lexHypo.getLastCoveredSrcStart();

            // I[_Ox̕
            // (Flex.hyp.ƂlastCoveredSrcEndقȂ̂ł̃^C~OłȂƃ_!!)
            ReorderingLabel reorderingLabel = ReorderingModel::getReorderingLabel(predBeg, predEnd, (int)start, (int)end);

            const float distPenalty  = ReorderingModel::calcDistortionPenalty(lexHypo.getLastCoveredSrcPos(), start);
            const float partialScore = lexHypo.getScore() - param.getWeightD()[DIST_PENALTY] * distPenalty;

            // future score estimation
            const float futureDistPenalty = (param.isFutureDistScoreUse() ? ReorderingModel::calcFutureDistortionPenalty(newHypoCoverage, cardinality, end) : 0.0F);
	    const float futureScore  = ltcManager.calcFutureScore(newHypoCoverage) - param.getWeightD()[DIST_PENALTY] * futureDistPenalty;

	    // őXRÃt[Y΂pruning
	    const PTableIndexPair& ptableIndexPair = ltcManager.getPhraseLattice().get(end, start);
	    const PhrasePairIndex& bestPP  = ptableIndexPair.first;
	    const PhrasePairIndex& worstPP = ptableIndexPair.second;
	    const float bestTMScore    =  bestPP->getScore();
	    const float unkWordPenalty = (bestPP != worstPP && bestPP->isUnkWord() ? (LanguageModel::unkWordPenalty*(end-start+1)) : 0.0F);

	    const float c0 = partialScore + bestTMScore + unkWordPenalty + futureScore;
	    if( isTooBad(c0, newCovHypo) )  continue;

            for(PhrasePairIndex iter = ptableIndexPair.first; iter!= ptableIndexPair.second; iter++){
              const PhrasePair& newAppliedPhrasePair = *iter;
              const float tmScore = newAppliedPhrasePair.getScore();

	      const float c1 = partialScore + tmScore + unkWordPenalty + futureScore;
              if( isTooBad(c1, newCovHypo) )  break;

              // I[_OfXRAvZ
              const float rmScore = computeReorderingScore(predPP, &newAppliedPhrasePair, reorderingLabel);
              const float c2 = c1 + rmScore;
              if( isTooBad(c2, newCovHypo) )  continue;

              // ꃂfXRǍvZ
	      const LMWordIDArray& newAppliedLMWordIDArray = newAppliedPhrasePair.getTgtPhrase().getLMWordIDArray();
              const LMWordIDArray& prevLMHistory = lexHypo.getLMHistory();
              const float lmScoreThreshold = (minScore - c2)/param.getWeightL();
              const float lmScore = lm.getLMScore(newAppliedLMWordIDArray, prevLMHistory/*, lmScoreThreshold*/);
              if(lmScore < lmScoreThreshold) continue;
              const float weightedLMScore = param.getWeightL() * lmScore;

              // XRǍvZ
              const float sentenceEndScore = (cardinality == sentenceLength) ? computeSentenceEndScore(newAppliedPhrasePair, end) : 0.0F;

	      float score = partialScore + tmScore + rmScore + unkWordPenalty + weightedLMScore + sentenceEndScore;

	      const float c3 = c2 + weightedLMScore;
              if( isTooBad(c3, newCovHypo) )  continue;
              if( maxScore < c3 ) updateMaxScore(c3);

              lexHypNum++;
	      LexicalHypothesis* newLexHyp = lexHypPool.pop();

	      if( !newCovHypo ){
		CoverageHypothesisMap::iterator newCovHypoIter = coverageHypoMap.insert( CoverageHypothesis(newHypoCoverage) ).first;
		newCovHypo = const_cast<CoverageHypothesis*>(&(*newCovHypoIter));
              }

	      newLexHyp->assign(
	        &newCovHypo->getCoverage(), score, futureScore,
	        &newAppliedPhrasePair,
	        lm.getLastHistory(), lm.getLastHistoryEnd(),
	        lm.getLastState(),
	        end, &lexHypo
              );

	      if( newCovHypo->push(newLexHyp) )
	        lexHypStack[lexHypNumOfCardinality++] = newLexHyp;
	      else
                lexHypPool.push(newLexHyp);

	      if(lexHypNumOfCardinality >= beamWidth*2)
	        pruneCardinality(lexHypStack, lexHypNumOfCardinality, coverageHypoMap);

            } // end of for(iter)
          } // end of for(lexHypoIndex)
        }// end of for(start)
      }// end of for(covHypoIndex)
    }// end of for(srcPhraseLength)

    pruneCardinalityLazy(cardinality, (const LexicalHypothesis**)lexHypStack, lexHypNumOfCardinality, coverageHypoMap);

  }// end of expandCardinality()

  void expandCoverage(){}

public:
  Translator(const LatticeManager& ltcManager)
   : sentenceLength(ltcManager.getPhraseLattice().size()),
     ltcManager(ltcManager),
     stackArray(sentenceLength+1),
     param(StaticData::getInstance().getParameter()),
     lm(StaticData::getInstance().getLanguageModel()),
     minScore(-20000.0F),
     maxScore(-10000.0F),
    beamThreshold(log(param.getBeamThreshold())),
    beamWidth(param.getBeamWidth())
     { initialize(); }


  const deque<const LexicalHypothesis*>& traceBack(const LexicalHypothesis* ptLexicalHypothesis) const {
    static deque<const LexicalHypothesis*> lexHypoPtrArray;
    lexHypoPtrArray.clear();

    for(const LexicalHypothesis* ptLexHypo = ptLexicalHypothesis;
         ptLexHypo->getPrevLexicalHypothesis() != NULL;
         ptLexHypo = ptLexHypo->getPrevLexicalHypothesis())
    {
      lexHypoPtrArray.push_front(ptLexHypo);
      if(StaticData::getInstance().getParameter().getVerbose() > 0) ptLexHypo->print();
    }// end of for

    return lexHypoPtrArray;
  }

  const deque<const LexicalHypothesis*>& traceBack() const {
    static deque<const LexicalHypothesis*> lexHypoPtrArray;
    lexHypoPtrArray.clear();

    for(const LexicalHypothesis* ptLexHypo = &stackArray.back().front()->getNthHypothesis(0);
         ptLexHypo->getPrevLexicalHypothesis() != NULL;
         ptLexHypo = ptLexHypo->getPrevLexicalHypothesis())
    {
      lexHypoPtrArray.push_front(ptLexHypo);
      if(StaticData::getInstance().getParameter().getVerbose() > 0)
        ptLexHypo->print();
    }// end of for

    return lexHypoPtrArray;
  }


  const string& translate(){
    // unsigned pruneCount[3] = {0};
    lexHypNum = 0;
    // CardinalityœWJ
    for(int cardinality = 1; cardinality <= sentenceLength; cardinality++){
      // cerr << "cardinality:" << cardinality << endl;
      expandCardinality(cardinality);

      // for debug
      if(param.getVerbose() > 1){
	const CoverageHypothesisStack& currentStack = stackArray[cardinality];
	cerr << "survived hypthesis of carkdinality " << cardinality << ": " << endl;
	for(unsigned i=0; i<currentStack.size(); i++)
          for(unsigned j=0; j<currentStack[i]->getStackSize(); j++)
	    cerr << (*currentStack[i])[j].getID() << "(" << (*currentStack[i])[j].getEstimatedScore() << ")" << endl;
	cerr << "cardinality" << cardinality << "end" << endl;
      }
    }

    // cerr << "pruned hyp: " << pruneCount[0] << ", " << pruneCount[1] << ", " << pruneCount[2] << endl;
    const LexicalHypothesis& bestHypothesis = stackArray.back().front()->getNthHypothesis(0);

    return convertHypoArrayToString( traceBack(&bestHypothesis) );
  }

  const float getBestScore() const {
    return stackArray.back().front()->getNthHypothesis(0).getEstimatedScore();
  }

  /*
  const string& getNthTranslation() const {
    return "hoge";
  }
  */

  unsigned getExpandedLexHypNum() const { return lexHypNum; }

};

#endif
