#ifndef COVERAGE_H_
#define COVERAGE_H_

#include <iostream>
#include <string>
#include "../etc/Typedef.h"
#include "../etc/Utility.h"

#define SENTENCE_LEN_MAX   (1024U)
#define UINT32_SIZE        (32U)
#define HASH_MAX           (10000U)
#define BITSET_ARRAY_SIZE  (SENTENCE_LEN_MAX/UINT32_SIZE)

#define MAKE_BIT_MASK(pos) (1U << (pos)%UINT32_SIZE)
#define CALC_BIT_IDX(pos)  ((pos) / UINT32_SIZE)

using namespace std;

struct Range{
  unsigned start;
  unsigned end;

  // Range(const unsigned s=0U, const unsigned e=0U): start(s), end(e){}

  const Range& operator=(const Range& rhs){
    this->start = rhs.start;
    this->end = rhs.end;
    return *this;
  }

  string toString() const {
    string str;
    str += "[";
    str += start;
    str += ", ";
    str += end;
    str += "]";
    return str;
  }
};

class Coverage{
private:
  static unsigned capacity;
  static unsigned size;

protected:
  unsigned bitset[BITSET_ARRAY_SIZE];

public:
  Coverage(){ initialize(); }

  Coverage(unsigned start, unsigned end){
    initialize();
    set(start, end);
  }

  Coverage(const Range& r){
    set(r);
  }

  Coverage(const Coverage& c){
    copy(&(c.bitset[0]), &(c.bitset[0])+capacity, &bitset[0]);
  }

  const Coverage& operator=(const Coverage& c){
    initialize();
    copy(&(c.bitset[0]), &(c.bitset[0])+capacity, &bitset[0]);
    return *this;
  }

  virtual ~Coverage(){}

  inline void initialize(){ memset(bitset, 0, capacity*UINT32_SIZE); }
  inline void clear(){ memset(bitset, 0, size); }

  inline void set(unsigned pos) {
    bitset[CALC_BIT_IDX(pos)] |= MAKE_BIT_MASK(pos);
  }

  void set(unsigned start, unsigned end) {
    for(unsigned i=start; i<=end; i++) set(i);
  }

  void set(const Range& r) {
    for(unsigned i=r.start; i<=r.end; i++) set(i);
  }

  inline void unset(unsigned pos) {
    bitset[CALC_BIT_IDX(pos)] &= ~MAKE_BIT_MASK(pos);
  }

  void reset(const Range& r, const bool value=false) {
    for(unsigned i=r.start; i<=r.end; i++) reset(i, value);
  }

  inline void reset(const unsigned pos, const bool value){
    value ? set(pos) : unset(pos);
  }

  inline bool get(unsigned pos) const {
    return bitset[CALC_BIT_IDX(pos)] & MAKE_BIT_MASK(pos);
  }

  inline void flip(unsigned pos) {
    bitset[CALC_BIT_IDX(pos)] ^= MAKE_BIT_MASK(pos);
  }


  bool getNextUncoveredRange(Range& out, unsigned pos=-1) const {
    // 0̈̎n_𒲂ׂ
    while(++pos < size)
      if(!get(pos)){ out.start = pos;  break; }

    if(pos == size) return false;

    // 0̈̏I_𒲂ׂ
    while(++pos < size)
      if(get(pos)) break;

    out.end = pos-1;
    return true;
  }

  string toString() const{
    string str;
    for(unsigned i=size; i>0; i--){
      if(i%8 == 0) str += ' ';
      str += (get(i-1) ? '1' : '0');
    }
    return str;
  }

  void print() const {
    for(unsigned i=0; i<size; i++){
      if(i%8 == 0) cerr << ' ';
      cerr << get(i);
    }
    cerr << endl;
  }

  bool operator==(const Coverage& rhs) const {
    for(unsigned i=0; i<=capacity; i++)
      if(bitset[i] != rhs.bitset[i]) return false;
    return true;
  }

  const Coverage& operator|=(const Coverage& rhs) {
    for(unsigned i=0; i<=capacity; i++)
      this->bitset[i] |= rhs.bitset[i];
    return *this;
  }

  const Coverage& operator|(const Coverage& rhs) const {
    static Coverage result;
    result = *this;
    result |= rhs;
    return result;
  }

  const Coverage& operator&=(const Coverage& rhs) {
    for(unsigned i=0; i<=capacity; i++)
      this->bitset[i] &= rhs.bitset[i];
    return *this;
  }

  const Coverage& operator&(const Coverage& rhs) const {
    static Coverage result;
    result = *this;
    result &= rhs;
    return result;
  }

  const Coverage& operator^=(const Coverage& rhs) {
    for(unsigned i=0; i<=capacity; i++)
      this->bitset[i] ^= rhs.bitset[i];
    return *this;
  }

  const Coverage& operator^(const Coverage& rhs) const {
    static Coverage result;
    result = *this;
    result ^= rhs;
    return result;
  }
/*
  bool operator<(const Coverage& rhs) const {
    return this < (&rhs);
  }
*/

  bool isOverlap(const Coverage& rhs) const {
    for(unsigned i=0; i<capacity; i++)
      if(bitset[i] & rhs.bitset[i]) return true;
    return false;
  }

  bool isOverlap(const Range& range) const {
    for(unsigned i=range.start; i<=range.end; i++)
      if(get(i)) return true;
    return false;
  }

  unsigned count(const Range& range) const {
    unsigned c = 0;
    for(unsigned i=range.start; i<=range.end; i++)
      if(get(i)) c++;
    return c;
  }

  unsigned count() const {
    unsigned c = 0;
    for(unsigned i=0; i<=size; i++)
      if(get(i)) c++;
    return c;
  }

  static void       setSize(const unsigned _size){
    size = _size;
    capacity = ((size+1)%UINT32_SIZE) ? (size+1)/UINT32_SIZE+1 : (size+1)/UINT32_SIZE;
  }
  static const unsigned getSize(){ return size; }

  bool operator<(const Coverage& rhs) const {
    for(int i=capacity-1; i>=0; i--)
      if(bitset[i] < rhs.bitset[i]) return true;
    return false;
  }

  unsigned getCount() const {
    unsigned count = 0U;
    for(unsigned i=0; i<size; i++)
      count += (unsigned)get(i);
    return count;
  }

  unsigned getUncovCount() const {
    return ( size-getCount() );
  }

  unsigned hash() const {
    unsigned h = 0;
    for(unsigned i=0; i<capacity; i++)
      h = h*7 + bitset[i];
    return h % HASH_MAX;
  }

};

#endif /* COVERAGE_H_ */
