#ifndef _LMODEL_
#define _LMODEL_

#include <Ngram.h>
#include <Vocab.h>
#include <File.h>
#include <LM.h>

#include <string>
#include <cmath>
#include <vector>
#include <deque>
#include <algorithm>

#include "../etc/Typedef.h"

using namespace std;

class Phrase;

class LanguageModel{
public:
  static const float   log10e;
  static const float   unkWordPenalty;
  static string        ssTag;
  static string        seTag;
  static LMWordID      ssTagID;
  static LMWordID      seTagID;
  static LMWordIDArray ssTagIDArray;
  static LMWordIDArray seTagIDArray;

  // historyp
  VocabIndex idArray[maxSentenceLength+1];
  unsigned  historySize;

  unsigned nGramOrder;     // ngram order
  Vocab*   vocab;            // Vocabrary
  Ngram*   ngramLM;          // N gram Language Model
  Boolean  limitVocab;

public:
  explicit LanguageModel(const unsigned& nGramOrder, const string& lmFileName)
    : nGramOrder(nGramOrder), ngramLM(NULL), vocab(NULL)
    {
      load(lmFileName.c_str());
      ssTagID = getID(ssTag);
      seTagID = getID(seTag);

      ssTagIDArray.push_back(ssTagID);
      seTagIDArray.push_back(seTagID);
    }

  virtual ~LanguageModel(){
    if(ngramLM) delete ngramLM;
    if(vocab)   delete vocab;
  }

  // Load Language Model File
  void load(const char* lmFileName){
    cerr << "======================== Load Language Model =======================" << endl;
    cerr << "Now Loading Language Model File..." << endl;
    vocab = new Vocab();
    limitVocab = false;
    vocab->unkIsWord() = true;  // consist unknown word
    vocab->toLower() = false;   // Not change Upper char to Lower char
    ngramLM = new Ngram(*vocab, nGramOrder); // create ngram LM
    File lmFile(lmFileName, "r");
    ngramLM->read(lmFile, limitVocab);
    if(StaticData::getInstance().getParameter().getVerbose() > 0){
      for(unsigned i=1; i<=nGramOrder; i++)
        cerr << "# " << i << "gram entry: " << ngramLM->numNgrams(i) << endl;
    }
    cerr << "Complete Loading Language Model File!" << endl;
    cerr << "====================================================================" << endl << endl;
  }

  // NgramRXgvZv~eBuwrapper֐
  inline float getNGramScore(const unsigned* ids) const{
    return (*ids != -1) ? (float)ngramLM->wordProb(*ids, ids+1)/log10e : unkWordPenalty;
  }

  // t[YꃂfRXǧvZ
  float getLMScore(const Phrase& phrase) ;

  // ɑ΂錾ꃂfRXǧvZ
  float getLMScoreForSentenceEnd();

  // ̌ꃂfRXǧvZ()
  float getLMScore(const LMWordIDArray& newPhrase, const LMWordIDArray& history = LMWordIDArray(), const float threshold = -1000.0F, bool debug = false) ;
  float getLMScore(LMWordIDArray& sentence, const unsigned sentenceSize, const float threshold = -1000.0F) const ;
  float getLMScore(const LMWordID* start, const LMWordID* end, const float threshold = -1000.0F) const {
    float score = 0.0F;
    for(const unsigned* pos = end-1; pos >= start; pos--){
      score += getNGramScore(pos);
// for debug
/*
      cerr << "p( " << getWord(*pos) << " | ";
      for(unsigned i=1; i<nGramOrder && (*(pos+i)) != Vocab_None; i++)
        cerr << getWord(*(pos+i)) << " ";
      cerr << ") = ";
      cerr << getNGramScore(pos) << endl;
*/
      if(score < threshold) break;
    }
    return score;
  }

  void              setNGramOrder(const unsigned& n) { nGramOrder = n; }

  // Pɑ΂錾ꃂf̎łID
  const LMWordID    getID(const string& w) const { return vocab->getIndex(w.c_str()); }
  // ꃂfIDP()擾
  const string      getWord(const LMWordID id) const {
    if(id != -1 && id != Vocab_None)
      return string(vocab->getWord(id));
    return "<unk>";
  }
  const unsigned    getNGramOrder()      const            { return nGramOrder; }
  const VocabIndex* getLastHistory()     const { return idArray; }
  const VocabIndex* getLastHistoryEnd()  const { return idArray + historySize; }
  const unsigned    getLastHistorySize() const { return historySize; }
  const void*       getLastState()       const {
    unsigned dummy;
    return ngramLM->contextID(idArray, dummy);
  }
};

#endif
