#ifndef LATTICEMANEGER_H_
#define LATTICEMANEGER_H_

#include <algorithm>

#include "../DataStructure/Phrase.h"
#include "../DataStructure/Sentence.h"
#include "../DataStructure/PhrasePair.h"
#include "../DataStructure/TriangleTable.h"

#include "../Model/PhraseTable.h"

#include "../DataStructure/Coverage.h"

using namespace std;

typedef TriangleTable<PTableIndexPair> PhraseLattice;
typedef TriangleTable<float>           FutureScoreLattice;
typedef TriangleTable<Coverage>        CoverageLattice;

class LatticeManager {
private:
  PhraseLattice     phraseLattice;
  FutureScoreLattice futureScoreLattice;
  CoverageLattice   coverageLattice;

  // t[YeBX̍쐬
  void buildPhraseLattice(const Sentence& srcSentence){
    const unsigned srcSentenceLength = srcSentence.size();
    const PhraseTable& phraseTable = StaticData::getInstance().getPhraseTable();

    for(unsigned start=0; start<srcSentenceLength; start++){
      unsigned tgtPhraseNum = 0;
      for(unsigned end=start; end<srcSentenceLength; end++){
        // t[Y
        const Phrase srcPhrase = srcSentence.getSubstring(start, end);
        // t[Ye[ũCfbNXۑpϐ
        PTableIndexPair pTableIndexPair = phraseTable.getUnkIndexPair();

        // t[Ỹt[Ye[uCfbNX擾
        if( phraseTable.isUnknownPhrase(srcPhrase) ){   // mt[Y̏ꍇ
          // młΎɒǉ
          pTableIndexPair = phraseTable.getUnkIndexPair();
          if( start == end && tgtPhraseNum == 0 ){
            pTableIndexPair = const_cast<PhraseTable&>(phraseTable).set( srcPhrase );
            tgtPhraseNum++;
          }
        }else{  // Ƀt[Y܂܂Ăꍇ
          pTableIndexPair = phraseTable.get(srcPhrase);
          if(distance(pTableIndexPair.first, pTableIndexPair.second) > StaticData::getInstance().getParameter().getTTableLimit()){

	    // pTableIndexPair.second = pTableIndexPair.second; //pTableIndexPair.first; // + StaticData::getInstance().getParameter().getTTableLimit();
	    // advance(pTableIndexPair.second, 20);
	    for(PhrasePairIndex idx = pTableIndexPair.first; idx != pTableIndexPair.second; idx++)
	      cerr << (*idx).getSrcPhrase().toString() << "<-> " << (*idx).getTgtPhrase().toString() << endl;
	  }
          tgtPhraseNum += distance(pTableIndexPair.first, pTableIndexPair.second);
        }

        phraseLattice.set( end, start, pTableIndexPair );
      }
    }
  }

  // CoverageeBX̍쐬
  void buildCoverageLattice() {
    const unsigned size = coverageLattice.size();
    for(unsigned start = 0; start<size; start++)
      for(unsigned end = start; end<size; end++){
        coverageLattice.get(end, start).set(start, end);
      }
  }

  // FutureScoreeBX̏
  void initFutureScoreLattice() {
    const unsigned srcSentenceLength = phraseLattice.size();
    for(unsigned start=0; start<srcSentenceLength; start++){
      float score = 0.0;
      for(unsigned end=start; end<srcSentenceLength; end++){
        score += (1.0F+StaticData::getInstance().getParameter().getWeightL()) * LanguageModel::unkWordPenalty;
        futureScoreLattice.set(end, start, score);
      } // end of for(start)
    } // end of for(end)
  }

  // FutureScoreeBX̍쐬
  void buildFutureScoreLattice() {
    const unsigned srcSentenceLength = phraseLattice.size();
    // Ԓω
    for(unsigned rangeSize=1; rangeSize<=srcSentenceLength; rangeSize++){
      // ԊJnω
      for(unsigned start=0; (start+rangeSize-1)<srcSentenceLength; start++){
        const unsigned end = start + (rangeSize-1);

        const PTableIndexPair& ptableIndexPair = phraseLattice.get(end, start);
        float bestScore;
	if(ptableIndexPair.first != ptableIndexPair.second )
	  bestScore = max_element(ptableIndexPair.first, ptableIndexPair.second)->getFutureScore();
	else
	  bestScore = (1.0F+StaticData::getInstance().getParameter().getWeightL()) * LanguageModel::unkWordPenalty * rangeSize;


        // _ω
        for(unsigned splitPos=start; splitPos<end; splitPos++){
          const float currentScore = futureScoreLattice.get(splitPos, start) + futureScoreLattice.get(end, splitPos+1);
          if(bestScore < currentScore) bestScore = currentScore;
        }// end of for(splitPos)

        // ŗǃXRAFutureScoreeBXɃZbg
        futureScoreLattice.set(end, start, bestScore);

      } // end of for(start)
    } // end of for(rangeSize)
  }  

public:
  LatticeManager(const Sentence& srcSentence)
  : phraseLattice(srcSentence.size()), futureScoreLattice(srcSentence.size()),
    coverageLattice(srcSentence.size())
  {
    buildPhraseLattice(srcSentence);
    buildCoverageLattice();
    buildFutureScoreLattice();
    if(StaticData::getInstance().getParameter().getVerbose() >= 1)  print();
  }

  const float calcFutureScore(const Coverage& coverage) const {
    static float    futureScore;
    static Coverage cash;
    if(coverage == cash)  return futureScore;
    else cash = coverage;

    futureScore = 0.0;

    Range r = {0, -1};
    while(coverage.getNextUncoveredRange(r, r.end))
      futureScore += futureScoreLattice.get(r.end, r.start);
    return futureScore;
  }

  void print() const {
    cerr << endl;
    cerr << "============= Phrase Table for Current Source Sentence =============" << endl;
    for(unsigned start=0; start<phraseLattice.size(); start++){
      for(unsigned end=start; end<phraseLattice.size(); end++){
        const PTableIndexPair& ptIndexPair = phraseLattice.get(end, start);
        if(StaticData::getInstance().getPhraseTable().getUnkIndexPair() == ptIndexPair) continue;
        cerr << "Source Phrase: " << (*ptIndexPair.first).getSrcPhrase().toString();
        cerr << " [" << start << "-" << end << "]";
	cerr << " (# Tgt Phrase: " << distance(ptIndexPair.first, ptIndexPair.second) << ")" << endl;
        unsigned count = 1;
        for(PhrasePairIndex index = ptIndexPair.first; index != ptIndexPair.second ; index++){
          const PhrasePair& phrasePair = (*index);
          cerr << "  " << count++  << "th:  ";
          cerr << phrasePair.getTgtPhrase().toString() << "\t ||| ";
          cerr << "future score: "<< phrasePair.getFutureScore() << ", ";
          cerr << "score: " << phrasePair.getScore() << endl ;
        } // end of for(index)
        cerr << endl;
      } // end of for(end)
    }// end of for(start)
    cerr << "====================================================================" << endl;
    cerr << endl;

    cerr << "========== Future Score Table for Current Source Sentence ===========" << endl;
    cerr << "start\\end";
    for(unsigned start=0; start<futureScoreLattice.size(); start++)
      fprintf(stderr, "    [%3u] ", start);
    cerr << endl;
    for(unsigned start=0; start<futureScoreLattice.size(); start++){
      fprintf(stderr, " [%3u]   ", start);
      for(unsigned end=0; end<futureScoreLattice.size(); end++){
        if(end<start)
          cerr << "          ";
        else
          fprintf(stderr, "%9.4f ", futureScoreLattice.get(end, start) );
      } // end of for(end)
      cerr << endl;
    }// end of for(start)
    cerr << "====================================================================" << endl;
    cerr << endl;
  }

  const PhraseLattice& getPhraseLattice() const {
    return phraseLattice;
  }

  const CoverageLattice& getCoverageLattice() const {
    return coverageLattice;
  }

  const FutureScoreLattice& getFutureScoreLattice() const {
    return futureScoreLattice;
  }

  virtual ~LatticeManager(){};
};

#endif /* LATTICEMANEGER_H_ */
