#ifndef _PHRASE_TABLE_
#define _PHRASE_TABLE_

#include <vector>
#include <deque>
#include <iostream>
#include <map>
#include <algorithm>
#include <string>
#include <utility>
#include <sstream>
#include <numeric>
#include <queue>


#include "../DataStructure/Phrase.h"
#include "../DataStructure/PhrasePair.h"
#include "../etc/Utility.h"

using namespace std;

// t[Ye[u֘Ǎ^ɂĂtypedef
typedef vector<PhrasePair>                     PhrasePairArray;
typedef PhrasePairArray::const_iterator        PhrasePairIndex;
// typedef const PhrasePair*                    PhrasePairIndex;
typedef pair<PhrasePairIndex, PhrasePairIndex>   PTableIndexPair;
typedef map<Phrase, PTableIndexPair>                 PhraseMap;
typedef pair<Phrase, PTableIndexPair>                PhraseMapEntry;
typedef deque<unsigned>                        SizeArray;


class PhraseTable{
 private:
  PhrasePairArray  phrasePairArray;
  PhraseMap        phraseMap;
  unsigned         ttableLimit;
  float            ttableThreshold;
  string           delimiter;
  SizeArray        tgtVariationNumArray;
  const Parameter* param;

  PhrasePairIndex  ssIndex;
  PhrasePairIndex  seIndex;

 private:
  void       printProgress(const unsigned currentNum, const unsigned tableSize) const {
    static unsigned prev;
    unsigned progress = (unsigned)( (float)currentNum/tableSize *100 );
    if(progress != 0 && progress > prev){
      cerr << ".";
      if( progress % 10 == 0 && progress % 100 != 0 )
	cerr << progress/10 << "0% ->" << endl;
      else if( progress % 100 == 0 )
	cerr << progress/10 << "0% //" << endl;
    }
    prev = progress;
  }

  PhrasePair strToPhrasePair(const string& phraseTableLine, const string& reorderingTableLine, const Parameter& param, LanguageModel& lm) const {
    /*
      t[Ye[ũtH[}bgi1sj:
         "src phrase ||| tgt phrase ||| src word align ||| tgt word align ||| feat.1 feat.2 feat.3 ..."

      ؂L = "|||"
    */

    // t[Ye[ûPsɊւ鏈
    StringArray ptableElems = split(phraseTableLine, param.getDelimiter());

    const string& ptableSrcPhrase = ptableElems[0];
    const string& ptableTgtPhrase = ptableElems[1];

    // skip word alignment(ptableElems[2,3]) in a phrase table line
    istringstream iStrStream(ptableElems[4]);
    const unsigned tmFeatureNum = param.getTMFeatureNum();
    ScoreArray tmScoreArray(tmFeatureNum);
    for(unsigned i=0; i<tmFeatureNum; i++)
      iStrStream >> tmScoreArray[i];
    tmScoreArray = log(tmScoreArray);
    tmScoreArray *= param.getWeightT();

    // ptableElems.pop_front();

    const unsigned tgtPhraseSize = split(ptableTgtPhrase).size();
    const float wordPenalty  = param.getWeightW() * (-(float)tgtPhraseSize);

    const float score        = accumulate(&tmScoreArray[0], &tmScoreArray[tmFeatureNum-1], wordPenalty); // inner_product(&tmScoreArray[0], &tmScoreArray[0]+tmFeatureNum, &param.getWeightT()[0], 0.0F) + wordPenalty;
    const float lmScore      = lm.getLMScore(Phrase(ptableTgtPhrase).getLMWordIDArray() );
    const float futureScore  = score + param.getWeightL()*lmScore;

    // I[_Oe[u1sɊւ鏈
    StringArray rtableElems = split(reorderingTableLine, param.getDelimiter());
    const string& rtableSrcPhrase = rtableElems[0];
    const string& rtableTgtPhrase = rtableElems[1];
    // I[_Oe[uƃt[Ye[ũGg̈v`FbN
    if(ptableSrcPhrase != rtableSrcPhrase && ptableTgtPhrase != rtableTgtPhrase){
      cerr << "Phrase Table Entry Not Matched Reordering Table Entry!!" << endl;
      cerr << "ptable entry: " << ptableSrcPhrase << " <-> " << ptableTgtPhrase << endl;
      cerr << "rtable entry: " << rtableSrcPhrase << " <-> " << rtableTgtPhrase << endl;
      exit(1);
    }

    istringstream rtableStrStream(rtableElems[2]);
    const unsigned rmFeatureNum = param.getRMFeatureNum();
    ScoreArray rmScoreArray(rmFeatureNum+1);
    rmScoreArray[DIST_PENALTY] = 1.0F;
    for(unsigned i=1; i<rmFeatureNum+1; i++)
      rtableStrStream >> rmScoreArray[i];
    rmScoreArray = log(rmScoreArray);
    rmScoreArray *= param.getWeightD();

    return PhrasePair(ptableSrcPhrase, ptableTgtPhrase, score, futureScore, tmScoreArray, rmScoreArray);
  }

  unsigned checkTTableSize(const string& ttableFileName) {
    ifstream ttableFile(ttableFileName.c_str());
    unsigned ttableSize = 0;
    unsigned allPhraseNum = 0;

    unsigned currentSrcPhraseNum = 0;
    string   currentSrcPhrase;
    string   line;

    cerr << "Check phrase table size" << endl;
    while( !ttableFile.eof() ){
      getline(ttableFile, line);
      if(line.empty()){
        if( !ttableFile.eof() )  cerr << "Caution: empty line exists in ttableFile, #line:" << allPhraseNum << endl;
        continue;
      }

      // i̕\
      allPhraseNum++;
      if( allPhraseNum   % (100*1000)  == 0 )
        cerr << ".";
      if( allPhraseNum   % (1000*1000) == 0 )
        cerr << allPhraseNum/(1000*1000) << "M" << endl;

      string srcPhrase = split(line, " "+delimiter+" ")[0];
      if(srcPhrase == currentSrcPhrase){
        if(currentSrcPhraseNum < ttableLimit)  currentSrcPhraseNum++;
	// tgtVariationNumArray.back()++;
      }else{
	// tgtVariationNumArray.push_back(1);
        ttableSize += currentSrcPhraseNum;
        currentSrcPhrase = srcPhrase;
        currentSrcPhraseNum = 1;
      }
    }
    cerr << endl;

    cerr << "# all  phrase: " << allPhraseNum << endl;
    cerr << "# load phrase: " << ttableSize;
    cerr << "(" << (float)ttableSize/allPhraseNum*100 << "%)" << endl << endl;

    ttableFile.close();
    return ttableSize;
  }

  void load(const Parameter& param, const string& ttableFileName, const string& rtableFileName, const unsigned ttableSize, LanguageModel& lm){
    this->param = &param;

    phrasePairArray.reserve((unsigned)(ttableSize*1.2));
    cerr << "reserved phrase table size: " << phrasePairArray.capacity() << endl;
    cerr << "load phrase table" << endl;
    ifstream ttableFile(ttableFileName.c_str());
    ifstream rtableFile(rtableFileName.c_str());
    typedef multiset<PhrasePair, greater<PhrasePair> > EquivarentPhraseSet;

    EquivarentPhraseSet equivPhraseSet;
    Phrase currentSrcPhrase;
    string ptableLine;
    string rtableLine;
    unsigned num = 0;

    while( !ttableFile.eof() || !rtableFile.eof() ){
      getline(ttableFile, ptableLine);
      getline(rtableFile, rtableLine);
      if( ptableLine.empty() || rtableLine.empty() ) continue;

      printProgress(num, ttableSize);

      PhrasePair phrasePair = strToPhrasePair(ptableLine, rtableLine, param, lm);

      // for debug
      if(StaticData::getInstance().getParameter().getVerbose() > 3)
        cerr << phrasePair << endl;

      const Phrase& src = phrasePair.getSrcPhrase();
      if(src.size() > maxPhraseLength) maxPhraseLength = src.size();

      if(src == currentSrcPhrase){
        equivPhraseSet.insert(phrasePair);
      }else{
        EquivarentPhraseSet::const_iterator start = equivPhraseSet.begin();
        EquivarentPhraseSet::const_iterator end   = equivPhraseSet.end();

	float threshold = -10000.0F;
	if(equivPhraseSet.size() > ttableLimit){
	  // check threshold
	  priority_queue<float> bestFutureScores;
	  for(EquivarentPhraseSet::const_iterator iter = start; iter != end; iter++)
	    bestFutureScores.push(iter->getFutureScore());
	  for(unsigned i=0; i<ttableLimit; i++)
	    bestFutureScores.pop();
	  threshold = bestFutureScores.top();
	}
	// push pp
	PhrasePairArray::const_iterator s = phrasePairArray.end();
	unsigned size = 0;
	for(EquivarentPhraseSet::const_iterator iter = start; size < ttableLimit && iter != end; iter++)
	  if(iter->getFutureScore() > threshold){
	    phrasePairArray.push_back(*iter);
	    size++;
	  }
	PhrasePairArray::const_iterator e = phrasePairArray.end();

        const PhraseMap::const_iterator it = phraseMap.insert( PhraseMapEntry(Phrase(currentSrcPhrase), PTableIndexPair(s, e)) ).first;

        num += distance(s, e);

        equivPhraseSet.clear();
        equivPhraseSet.insert(phrasePair);
        currentSrcPhrase = src;
      }

    }// end of while

    ssIndex = setSSPhrasePair().first;
    seIndex = setSEPhrasePair().first;

    printProgress(num, ttableSize);
    cerr << "# loaded phrase: " << num << endl;
    cerr << "complete loading phrase table!!"  << endl;
    cerr << endl;
  }

 public:
  explicit PhraseTable(const Parameter& param, LanguageModel& lm)
    :ttableLimit(param.getTTableLimit()), ttableThreshold( param.getTTableThreshold() ), delimiter(param.getDelimiter())
  {
    const string& ttableFileName = param.getTTableFileName();
    const string& rtableFileName = param.getRTableFileName();
    unsigned ttableSize = checkTTableSize(ttableFileName);
    load(param, ttableFileName, rtableFileName, ttableSize, lm);

  }

  bool isUnknownPhrase(const Phrase& srcPhrase) const {
    return phraseMap.end() == phraseMap.find(srcPhrase);
  }

  const PTableIndexPair& set(const Phrase& srcPhrase) { // srcPhrase:m
    const string phrase = strip(srcPhrase.toString());
    const ScoreArray tmScoreArray(0.0, StaticData::getInstance().getParameter().getTMFeatureNum());

    ScoreArray rmScoreArray(log(0.333333F), StaticData::getInstance().getParameter().getRMFeatureNum()+1); // vC: reorderingfƂ
    rmScoreArray[DIST_PENALTY] = 0.0F;
    rmScoreArray *= StaticData::getInstance().getParameter().getWeightD();

    // mt[Y΂ƂĐVɒ`
    const PhrasePair newPhrasePair(phrase, phrase, (float)0.0, (1.0F+param->getWeightL())*LanguageModel::unkWordPenalty, tmScoreArray, rmScoreArray);

    const PhrasePairIndex start= phrasePairArray.end();
    // t[Y΂̎̂ǉ
    phrasePairArray.push_back(newPhrasePair);
    const PhrasePairIndex end = phrasePairArray.end();

    // t[Y΂̃CfbNXɒǉ
    const PTableIndexPair indexPair = PTableIndexPair(start, end);
    return ( *( phraseMap.insert( PhraseMapEntry(srcPhrase, indexPair) ).first ) ).second;
  }

  const PTableIndexPair& get(const Phrase& srcPhrase) const {
    return (*phraseMap.find(srcPhrase)).second;
  }


  const PTableIndexPair getUnkIndexPair() const {
    return PTableIndexPair(phrasePairArray.end(), phrasePairArray.end());
  }

  const PTableIndexPair& setSSPhrasePair() {
    const string phrase = LanguageModel::ssTag;
    const ScoreArray tmScoreArray(0.0, StaticData::getInstance().getParameter().getTMFeatureNum());

    ScoreArray rmScoreArray(0.0F, StaticData::getInstance().getParameter().getRMFeatureNum()+1);

    // mt[Y΂ƂĐVɒ`
    const PhrasePair newPhrasePair(phrase, phrase, 0.0F, 0.0F, tmScoreArray, rmScoreArray);

    const PhrasePairIndex start= phrasePairArray.end();
    // t[Y΂̎̂ǉ
    phrasePairArray.push_back(newPhrasePair);
    const PhrasePairIndex end = phrasePairArray.end();

    // t[Y΂̃CfbNXɒǉ
    const PTableIndexPair indexPair = PTableIndexPair(start, end);
    return ( *( phraseMap.insert( PhraseMapEntry(Phrase(phrase), indexPair) ).first ) ).second;
  }

  const PTableIndexPair& setSEPhrasePair() {
    const string phrase = LanguageModel::seTag;
    const ScoreArray tmScoreArray(0.0, StaticData::getInstance().getParameter().getTMFeatureNum());

    ScoreArray rmScoreArray(0.0F, StaticData::getInstance().getParameter().getRMFeatureNum()+1);

    // mt[Y΂ƂĐVɒ`
    const PhrasePair newPhrasePair(phrase, phrase, 0.0F, 0.0F, tmScoreArray, rmScoreArray);

    const PhrasePairIndex start= phrasePairArray.end();
    // t[Y΂̎̂ǉ
    phrasePairArray.push_back(newPhrasePair);
    const PhrasePairIndex end = phrasePairArray.end();

    // t[Y΂̃CfbNXɒǉ
    const PTableIndexPair indexPair = PTableIndexPair(start, end);
    return ( *( phraseMap.insert( PhraseMapEntry(Phrase(phrase), indexPair) ).first ) ).second;
  }

  const PhrasePairIndex& getSSPhrasePair() const { return ssIndex; }
  const PhrasePairIndex& getSEPhrasePair() const { return seIndex; }

  void print() const {
    for(unsigned i=0; i<phrasePairArray.size(); i++)
      cerr << phrasePairArray[i].toString() << endl;
  }
};

#endif
