#ifndef _REORDERING_MODEL_
#define _REORDERING_MODEL_

#include <iostream>
#include <deque>
#include <string>
#include <cmath>

#include "../etc/Utility.h"
#include "../DataStructure/Coverage.h"
#include "../DataStructure/PhrasePair.h"

using namespace std;

enum Type           {monotonic, msd, none};
enum Direction      {uni, bi};
enum Identification {f, fe};

class ReorderingType{
 public:
  Type           type;
  Direction      dir;
  Identification id;

 public:
  ReorderingType(): type(none), dir(uni) {}


  void set(const string& reorderingType){
    StringArray elems = split(reorderingType, "-");

    for(unsigned i=0; i<elems.size(); i++){
      string elem = elems[i];
      if (elem == "msd")                type = msd;
      else if (elem == "monotonicity")  type = monotonic;
      else if (elem == "bidirectional") dir  = bi;
      else if (elem == "f")             id   = f;
      else if (elem == "fe")            id   = fe;
    } // end of for
  }

  string toString() const {
    string strType;
    // reordering type
    switch(type){
    case monotonic:
      strType += "monotonicity-";  break;
    case msd:
      strType += "msd-";        break;
    } // end of switch

    // direction
    switch(dir){
    case uni:
      break;
    case bi:
      strType += "bidirectional-";  break;
    } // end of switch

    // identification
    switch(id){
    case f:
      strType += "f";    break;
    case fe:
      strType += "fe";   break;
    } // end of switch
    return strType;
  }

  void print() const {
    // reordering type
    switch(type){
    case monotonic:
      cerr << "monotonicity-";  break;
    case msd:
      cerr << "msd-";           break;
    } // end of switch

    // direction
    switch(dir){
    case uni:
      break;
    case bi:
      cerr << "bidirectional-";  break;
    } // end of switch

    // identification
    switch(id){
    case f:
      cerr << "f";    break;
    case fe:
      cerr << "fe";   break;
    } // end of switch

    cerr << "& distance" << endl;
  }
};

enum ReorderingLabel{
  MONO_LABEL = 0,
  SWAP_LABEL,
  DISCONT_LABEL,
  LABEL_NUM
};

static string ReorderingLabelName[LABEL_NUM] = {
    "monotone",
    "swap",
    "discont."
};

enum MsdBidirectionalFE{
  DIST_PENALTY = 0,
  P_MONO,   // P: Predecessor
  P_SWAP,
  P_DISCONT,
  S_MONO,       // S: Successor
  S_SWAP,
  S_DISCONT,
  FEATURE_NUM
};

static string FeatureNameMSDBFE[FEATURE_NUM] = {
    "penalty",
    "mono(p)",
    "swap(p)",
    "discont.(p)",
    "mono(s)",
    "swap(s)",
    "discont.(s)"
};


class ReorderingModel{
public:
  static const float getLBOScore(const PhrasePair* const prev, const PhrasePair* const post, const ReorderingLabel label) {
    float score = 0.0F;
    switch(label){
    case MONO_LABEL:
      score += prev->getRMScore(S_MONO);
      score += post->getRMScore(P_MONO);
      break;
    case SWAP_LABEL:
      score += prev->getRMScore(S_SWAP);
      score += post->getRMScore(P_SWAP);
      break;
    default:
      score += prev->getRMScore(S_DISCONT);
      score += post->getRMScore(P_DISCONT);
    }

    return score;
  }

  static const ReorderingLabel getReorderingLabel(const int preBeg, const int preEnd, const int sucBeg, const int sucEnd) {
    ReorderingLabel label = DISCONT_LABEL;

    if(sucBeg == preEnd+1) label = MONO_LABEL;
    else if(preBeg == sucEnd+1) label = SWAP_LABEL;

    return label;
  }

  static const ReorderingLabel getReorderingLabel(const Range& prev, const Range& post) {
    const int preBeg = (int)prev.start;
    const int preEnd = (int)prev.end;
    const int sucBeg = (int)post.start;
    const int sucEnd = (int)post.end;

    ReorderingLabel label = DISCONT_LABEL;

    if(sucBeg == preEnd+1) label = MONO_LABEL;
    else if(preBeg == sucEnd+1) label = SWAP_LABEL;

    return label;
  }


  static const float calcFutureDistortionPenalty(const Coverage& coverage, const unsigned cardinality, int lastCoveredEnd) {
    Range r = {0, -1};
    float futureDistPenalty = 0.0F;
    
    while( !coverage.getNextUncoveredRange(r, r.end) ){
      futureDistPenalty += (float)abs(lastCoveredEnd - (int)r.start + 1);
      lastCoveredEnd  = (int)r.end;
      r.end++;
    }
    futureDistPenalty += abs(lastCoveredEnd - (int)Coverage::getSize() + 1);

    return futureDistPenalty;
  }
  /*
  static const float calcDistortionPenalty2(const Coverage& coverage, const unsigned cardinality, int lastCoveredEnd) {
    Range r = {0,-1};
    if( !coverage.getNextUncoveredRange(r, r.end) ) return 0.0F;

    const unsigned firstUncoveredPosition = r.start;
    const int distortionOfNextJump = abs((int)lastCoveredEnd - (int)firstUncoveredPosition + 1);
    const int estimatedJumpTimes   = (int)cardinality - (int)firstUncoveredPosition + 1;
    
    return (float)(estimatedJumpTimes + distortionOfNextJump);
  }
  */
  static const float calcDistortionPenalty(const int lastPhraseEnd, const int newPhraseStart) {
    return (float)abs(lastPhraseEnd - newPhraseStart + 1);
  }
};

#endif
