#ifndef _PARAMETER_
#define _PARAMETER_

#include <iostream>
#include <fstream>
#include <vector>
#include <valarray>
#include <string>
#include <cmath>

#include "../Model/ReorderingModel.h"

using namespace std;

typedef valarray<float> ScoreArray;
typedef valarray<float> WeightArray;

enum SeedType{
  BEAM_SEARCH,
  SCRATCH,
  VITERBI
};

class Parameter{
 private:
  // input(source language) file & output(result) file
  string      inputFileName;
  string      outputFileName;
  string      initFileName;

  // parameters for translation model
  string      ttableFileName;
  unsigned    ttableLimit;
  float       ttableThreshold;
  unsigned    tmSrcFactorNum;
  unsigned    tmRefFactorNum;
  unsigned    tmFeatureNum;    // number of weight for translation model
  WeightArray weightT;

  // parameters for language model
  string      lmFileName;
  unsigned    lmFactorNum;
  unsigned    lmFormatType;
  unsigned    ngramOrder;
  float       weightL;

  // parameters for reordering(distortion) model
  string         rtableFileName;
  int            distortionLimit;
  unsigned       rmSrcFactorNum;
  unsigned       rmRefFactorNum;
  unsigned       rmFeatureNum;    // number of weight for reordering model
  ReorderingType rmType;
  WeightArray    weightD;

  // parameter for word penalty
  float       weightW;
  


  // parameters for decording
  unsigned    cardinalityStackSize;
  float       beamThreshold;
  bool        futureDistScoreFlag;
  /*
  unsigned    cardinalityStackSizeForCoverage;
  unsigned    coverageStackSize;
  */

  // delimiter of phrase table
  string      delimiter;

  // optional parameter
  unsigned    verbose;


 private:

 public:
  Parameter() :tmFeatureNum(5U), weightT(0.2F, tmFeatureNum), rmFeatureNum(0U), weightD(0.3F, rmFeatureNum), weightW(0.5F), weightL(0.5F),
  distortionLimit(-1), verbose(0U),
  cardinalityStackSize(256U), beamThreshold(0.00001F), futureDistScoreFlag(false)
  {}

  Parameter(const string initFileName){
    readInitFile(initFileName);
  }

  void usage() const {
    cerr << "usage: \% khafra++ -f ConfigFile [options...]" << endl;
    cerr << "options: " << endl;

    // list of khafra++ options
    cerr << "\t -config(-f)  config_file : Configration File Path" << endl;
    cerr << "\t -verbose(-v) level       : Verbose Level(default=0)" << endl;
    cerr << "                              level=1 : Phrase_List and Lattice for Each Src Sentence" << endl;
    cerr << "                              level=2 : + Hypothesis_List" << endl;
    cerr << "                              level=3 : + PhraseTable" << endl;
    cerr << "                              level>3 : + more information" << endl;
    cerr << "\t -stack(-s)   stack_size  : Stack Size for Each Cardinality" << endl;
    cerr << "\t -beam-threshokd(-b) thre : Beam Width for Each Cardinality" << endl;
    cerr << "\t -use-future-dist-cost    : Using Future Distortion Score for Rest Score Estimation" << endl;

  }

  void print() const {
    cerr << "================================ Parameters =================================" << endl;
    cerr << "[Configuration File]" << endl;
    cerr << "  " << initFileName << endl;
    cerr << "[Translation Model]" << endl;
    cerr << "  # Src Factor: " << tmSrcFactorNum << endl;
    cerr << "  # Tgt Factor: " << rmRefFactorNum << endl;
    cerr << "  # TM Feature: " << tmFeatureNum << endl;
    cerr << "  PTable File : " << ttableFileName << endl;

    cerr << "[Delimiter]" << endl;
    cerr << "  delimiter: " << delimiter << endl;

    cerr << "[Language Model]" << endl;
    cerr << "  Format     : " << (lmFormatType ? "isrtlm" : "srilm") << endl;
    cerr << "  # Factor   : " << lmFactorNum << endl;
    cerr << "  Ngram Order: " << ngramOrder << endl;
    cerr << "  LM File    : " << lmFileName << endl;

    cerr << "[Reordering Model]" << endl;
    cerr << "  Distortion Limit: " << distortionLimit << endl;
    cerr << "  # Src Factor    : " << rmSrcFactorNum << endl;
    cerr << "  # Ref Factor    : " << rmRefFactorNum << endl;
    cerr << "  # RM Feature    : " << rmFeatureNum  << endl;
    cerr << "  Reordering Type : " << rmType.toString() << endl;
    cerr << "  RTable File     : " << rtableFileName << endl;

    cerr << "[Weight for Features]" << endl;
    cerr << "  weight-d:";
    for(unsigned i=0; i<rmFeatureNum+1; i++)
      cerr << " " << weightD[i];
    cerr << endl;
    cerr << "  weight-t:";
    for(unsigned i=0; i<weightT.size(); i++)
      cerr << " " << weightT[i];
    cerr << endl;
    cerr << "  weight-l: " << weightL << endl;
    cerr << "  weight-w: " << weightW << endl;

    cerr << "[Decoding Parameters]" << endl;
    cerr << "  Stack Size     : " << cardinalityStackSize << endl;
    cerr << "  Beam Threshold : " << beamThreshold << endl;
    cerr << "  Rest Dist Score: " << ((futureDistScoreFlag) ? "on" : "off") << endl;
    cerr << "=============================================================================" << endl;
    cerr << endl;
}

  static Parameter* readArgs(const unsigned argc, const char** argv){
    Parameter* ptParam = new Parameter();
    unsigned argNum = 1;

    if( !strcmp("-f", argv[argNum]) || !strcmp("-config", argv[argNum])){
      string initFileName = argv[++argNum];
      ptParam->readInitFile(initFileName);
    }else{
      ptParam->usage();
      exit(1);
    }

    // TO DO: R}hC̉
    while(++argNum < argc){
      if( !strcmp("-verbose", argv[argNum]) || !strcmp("-v", argv[argNum]) )
        ptParam->setVerbose(atoi(argv[++argNum]));
      else if ( !strcmp("-input", argv[argNum]) || !strcmp("-i", argv[argNum]))
        ptParam->setInputFileName(argv[++argNum]);
      else if ( !strcmp("-ttable-limit", argv[argNum]) || !strcmp("-ttl", argv[argNum]))
        ptParam->setTTableLimit((unsigned)atoi(argv[++argNum]));
      else if ( !strcmp("-stack", argv[argNum]) || !strcmp("-s", argv[argNum]) )
	ptParam->setStackSize((unsigned)atoi(argv[++argNum]));
      else if ( !strcmp("-beam-threshold", argv[argNum]) || !strcmp("-b", argv[argNum]) )
	ptParam->setBeamThreshold((float)atof(argv[++argNum]));
      else if ( !strcmp("-use-future-dist-cost", argv[argNum]) )
	ptParam->setUseFutureDistScore();
    } // end of while
    ptParam->print();
    return ptParam;
  }

  void readInitFile(const string initFileName){
    ifstream initFile(initFileName.c_str() );
    if( initFile.fail() ){
      cerr << "can't open configuration file: " << initFileName << " !!" << endl;
      exit(1);
    }
    this->initFileName = initFileName;

    string line;
    while(!initFile.eof()){
      getline(initFile, line);
      if(line == "[ttable-limit]"){
        initFile >> ttableLimit;
        ifstream::pos_type pos = initFile.tellg();
        getline(initFile, line);// initFile >> ttableThreshold;
        for(int i=0; i<line.size(); i++)
          if( !(('0' <= line[i] && line[i] <= '9') || line[i] == '.') ){
            initFile.seekg(pos);
            line = "0.0001";
            break;
          }
        ttableThreshold = (float)atof(line.c_str());
        ttableThreshold = log(ttableThreshold);
      } else if (line == "[ttable-file]") {
        initFile >> tmSrcFactorNum;
        initFile >> tmRefFactorNum;
        initFile >> tmFeatureNum;
        initFile >> ttableFileName;
      } else if (line == "[distortion-file]") {
        string temp;
        initFile >> temp;
        sscanf(temp.c_str(), "%u-%u", &rmSrcFactorNum, &rmRefFactorNum);
        initFile >> temp;
        rmType.set(temp);
        initFile >> rmFeatureNum;
        weightD.resize(rmFeatureNum+1, 0.0F);
        initFile >> rtableFileName;
      } else if (line == "[lmodel-file]") {
        initFile >> lmFormatType;
        initFile >> lmFactorNum;
        initFile >> ngramOrder;
        initFile >> lmFileName;
      } else if (line == "[weight-d]") {
        for(unsigned i=0; i<rmFeatureNum+1; i++)
          initFile >> weightD[i];
      } else if (line == "[weight-t]") {
        for(unsigned i=0; i<weightT.size(); i++)
          initFile >> weightT[i];
      } else if (line == "[weight-l]" ) {
        initFile >> weightL;
      } else if (line == "[weight-w]") {
        initFile >> weightW;
      } else if (line == "[distortion-limit]") {
        initFile >> distortionLimit;
      } else if (line == "[factor-delimiter]") {
        initFile >> delimiter;
      } else if (line == "[stack-size]") {
        initFile >> cardinalityStackSize;
      }// end of switch(line)
    }// end of while
  }

  // setter
  void setVerbose(const unsigned level)     { verbose = level; }
  void setInputFileName(const string fname) { inputFileName = fname; }
  void setTTableLimit(const unsigned ttl)   { ttableLimit = ttl; }
  void setStackSize(const unsigned size)    { cardinalityStackSize = size; }
  void setBeamThreshold(const float thre)   { beamThreshold = thre; }
  void setUseFutureDistScore(bool flag=true) { futureDistScoreFlag = flag; }

  // getter
  unsigned           getVerbose()        const { return verbose; }

  const string&      getLMFileName()     const { return lmFileName; }
  unsigned           getNGramOrder()     const { return ngramOrder; }
  const float        getWeightL()        const { return weightL; }

  const WeightArray& getWeightD()        const { return weightD; }
  const float        getWeightW()        const { return weightW; }

  const string&      getTTableFileName() const { return ttableFileName; }
  unsigned           getTTableLimit()    const { return ttableLimit; }
  const float        getTTableThreshold() const { return ttableThreshold; }

  unsigned           getTMFeatureNum()   const { return tmFeatureNum; }
  const WeightArray& getWeightT()        const { return weightT; }

  const string&      getRTableFileName() const { return rtableFileName; }
  const unsigned     getRMFeatureNum()   const { return rmFeatureNum; }


  const string&      getDelimiter()      const { return delimiter; }

  const string&      getInputFileName()  const { return inputFileName; }

  const unsigned     getBeamWidth()      const { return cardinalityStackSize; }
  const float        getBeamThreshold()  const { return beamThreshold; }
  const bool         isFutureDistScoreUse() const { return futureDistScoreFlag; }
};

#endif
